//
// goamz - Go package to interact with Linux Containers (LXC).
//
// https://wiki.ubuntu.com/golxc
//
// Copyright (c) 2012 Canonical Ltd.
//
// Written by Aram Hăvărneanu <aram.havarneanu@canonical.com>
// and Frank Mueller <frank.mueller@canonical.com>
//

package golxc

import (
	"bytes"
	"fmt"
	"log"
	"os"
	"os/exec"
	"strconv"
	"strings"
)

const (
	// ContainerStopped represents a stopped container.
	ContainerStopped = 1 << iota

	// ContainerStarting represents a starting container.
	ContainerStarting = 1 << iota

	// ContainerRunning represents a running container.
	ContainerRunning = 1 << iota

	// ContainerAborting represents a starting container after
	// an error occured.
	ContainerAborting = 1 << iota

	// ContainerStopping represents a stopping container.
	ContainerStopping = 1 << iota
)

// state2go maps lxc state strings to the according go constants.
var state2go = map[string]int{
	"STOPPED":  ContainerStopped,
	"STARTING": ContainerStarting,
	"RUNNING":  ContainerRunning,
	"ABORTING": ContainerAborting,
	"STOPPING": ContainerStopping,
}

// go2state maps go state constants to the according lxc strings.
var go2state = map[int]string{
	ContainerStopped:  "STOPPED",
	ContainerStarting: "STARTING",
	ContainerRunning:  "RUNNING",
	ContainerAborting: "ABORTING",
	ContainerStopping: "STOPPING",
}

const (
	LogDebug = iota
	LogInfo
	LogNotice
	LogWarning
	LogError
	LogCritical
	LogFatal
)

// log2go maps the lxc log strings to the according go constants.
var log2go = map[string]int{
	"DEBUG":  LogDebug,
	"INFO":   LogInfo,
	"NOTICE": LogNotice,
	"WARN":   LogWarning,
	"ERROR":  LogError,
	"CRIT":   LogCritical,
	"FATAL":  LogFatal,
}

// go2log maps the go log constants to the according lxc strings.
var go2log = map[int]string{
	LogDebug:    "DEBUG",
	LogInfo:     "INFO",
	LogNotice:   "NOTICE",
	LogWarning:  "WARN",
	LogError:    "ERROR",
	LogCritical: "CRIT",
	LogFatal:    "FATAL",
}

// Container represents a linux container instance and provides
// operations to create, maintain and destroy the container. 
type Container struct {
	name     string
	logFile  string
	logLevel string
}

// New returns a container instance which can then be used for operations
// like Create(), Start(), Stop() or Destroy().
func New(name string) *Container {
	c := &Container{
		name: name,
	}
	// Debugging infos.
	state, pid, err := c.Info()
	if err != nil {
		log.Fatalf("cannot retrieve container info for %q: %v", c.name, err)
		return nil
	}
	log.Printf("container %q has state %q and pid %d", c.name, go2state[state], pid)
	return c
}

// List lists the existing containers on the system.
func List() ([]*Container, error) {
	args := []string{"-1"}
	cmd := exec.Command("lxc-ls", args...)
	outbuf, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot list containers: %v", err)
		return nil, err
	}
	names := nameSet(outbuf)
	containers := []*Container{}
	for _, name := range names {
		containers = append(containers, New(name))
	}
	return containers, nil
}

// SetLog configures the logging for the lxc commands.
func (c *Container) SetLog(logFile string, logLevel int) {
	c.logFile = logFile
	c.logLevel = go2log[logLevel]
}

// Create creates a new container based on the given template.
func (c *Container) Create(template string, tmplArgs ...string) error {
	if c.IsConstructed() {
		return nil
	}
	args := []string{
		"-n", c.name,
		"-t", template,
	}
	if len(tmplArgs) != 0 {
		// Arguments after -- will be passed to the template.
		args = append(append(args, "--"), tmplArgs...)
	}
	cmd := exec.Command("lxc-create", args...)
	_, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot create container %q: %v", c.name, err)
		return err
	}
	return nil
}

// Start runs the container as a daemon.
func (c *Container) Start(configFile, consoleFile string) error {
	args := []string{
		"--daemon",
		"-n", c.name,
	}
	if configFile != "" {
		args = append(args, "-f", configFile)
	}
	if consoleFile != "" {
		args = append(args, "-c", consoleFile)
	}
	if c.logFile != "" {
		args = append(args, "-o", c.logFile, "-l", c.logLevel)
	}
	cmd := exec.Command("lxc-start", args...)
	_, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot start container %q: %v", c.name, err)
		return err
	}
	return c.wait(ContainerRunning)
}

// Stop terminates the running container.
func (c *Container) Stop() error {
	args := []string{
		"-n", c.name,
	}
	if c.logFile != "" {
		args = append(args, "-o", c.logFile, "-l", c.logLevel)
	}
	cmd := exec.Command("lxc-stop", args...)
	_, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot stop container %q: %v", c.name, err)
		return err
	}
	return c.wait(ContainerStopped)
}

// Clone creates a copy of the container, it gets the given name.
func (c *Container) Clone(name string) (*Container, error) {
	if !c.IsConstructed() {
		return nil, fmt.Errorf("container %q is not yet created", c.name)
	}
	cc := &Container{
		name: name,
	}
	if cc.IsConstructed() {
		return cc, nil
	}
	args := []string{
		"-o", c.name,
		"-n", name,
	}
	cmd := exec.Command("lxc-clone", args...)
	_, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot clone container %q: %v", c.name, err)
		return nil, err
	}
	return cc, nil
}

// Freeze freezes all the container'ss processes.
func (c *Container) Freeze() error {
	if !c.IsConstructed() {
		return fmt.Errorf("container %q is not yet created", c.name)
	}
	args := []string{
		"-n", c.name,
	}
	if c.logFile != "" {
		args = append(args, "-o", c.logFile, "-l", c.logLevel)
	}
	cmd := exec.Command("lxc-freeze", args...)
	_, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot freeze container %q: %v", c.name, err)
		return err
	}
	return nil
}

// Unfreeze thaws all a froozen container's processes.
func (c *Container) Unfreeze() error {
	if !c.IsConstructed() {
		return fmt.Errorf("container %q is not yet created", c.name)
	}
	args := []string{
		"-n", c.name,
	}
	if c.logFile != "" {
		args = append(args, "-o", c.logFile, "-l", c.logLevel)
	}
	cmd := exec.Command("lxc-unfreeze", args...)
	_, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot unfreeze container %q: %v", c.name, err)
		return err
	}
	return nil
}

// Destroy stops and removes the container.
func (c *Container) Destroy() error {
	if !c.IsConstructed() {
		return fmt.Errorf("container %q is not yet created", c.name)
	}
	if err := c.Stop(); err != nil {
		return err
	}
	args := []string{
		"-n", c.name,
	}
	cmd := exec.Command("lxc-destroy", args...)
	_, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot destroy container %q: %v", c.name, err)
		return err
	}
	return nil
}

// Info returns the status and the process id of the container.
func (c *Container) Info() (int, int, error) {
	args := []string{
		"-n", c.name,
	}
	cmd := exec.Command("lxc-info", args...)
	outbuf, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot retrieve container info for %q: %v", c.name, err)
		return -1, -1, err
	}
	kv := keyValues(outbuf, ": ")
	state := state2go[kv["state"]]
	pid, err := strconv.Atoi(kv["pid"])
	if err != nil {
		return -1, -1, err
	}
	return state, pid, nil
}

// IsConstructed checks if the container image exists.
func (c *Container) IsConstructed() bool {
	fi, err := os.Stat(c.rootfs())
	if err != nil {
		return false
	}
	return fi.IsDir()
}

// IsRunning checks if the state of the container is 'RUNNING'.
func (c *Container) IsRunning() bool {
	state, _, err := c.Info()
	if err != nil {
		return false
	}
	return state == ContainerRunning
}

// containerHome returns the name of the container directory.
func (c *Container) containerHome() string {
	return "/var/lib/lxc/" + c.name
}

// rootfs returns the name of the directory containing the
// root filesystem of the container.
func (c *Container) rootfs() string {
	return c.containerHome() + "/rootfs/"
}

// wait waits for the passed container status.
func (c *Container) wait(status int) error {
	lxcStatus := []string{}
	if status&ContainerStarting != 0 {
		lxcStatus = append(lxcStatus, "STARTING")
	}
	if status&ContainerRunning != 0 {
		lxcStatus = append(lxcStatus, "RUNNING")
	}
	if status&ContainerAborting != 0 {
		lxcStatus = append(lxcStatus, "ABORTING")
	}
	if status&ContainerStopping != 0 {
		lxcStatus = append(lxcStatus, "STOPPING")
	}
	if status&ContainerStopped != 0 {
		lxcStatus = append(lxcStatus, "STOPPED")
	}
	lxcStatusStr := strings.Join(lxcStatus, "|")
	args := []string{
		"-n", c.name,
		"-s", lxcStatusStr,
	}
	cmd := exec.Command("lxc-wait", args...)
	_, _, err := run(cmd)
	if err != nil {
		log.Fatalf("cannot wait for container %q: %v", c.name, err)
		return err
	}
	return nil
}

// run executes the passed command and returns stdout and stderr.
func run(cmd *exec.Cmd) (outbuf, errbuf bytes.Buffer, err error) {
	cmd.Stdout = &outbuf
	cmd.Stderr = &errbuf
	err = cmd.Run()
	if err != nil {
		log.Fatalf("error executing %q: %v", strings.Join(cmd.Args, " "), err)
		return
	}
	return
}

// keyValues retrieves key/value pairs out of a command output.
func keyValues(raw bytes.Buffer, sep string) map[string]string {
	kv := map[string]string{}
	lines := strings.Split(raw.String(), "\n")
	for _, line := range lines {
		parts := strings.SplitN(line, sep, 2)
		if len(parts) == 2 {
			kv[parts[0]] = strings.TrimSpace(parts[1])
		}
	}
	return kv
}

// nameSet retrieves a set of names out of a command output.
func nameSet(raw bytes.Buffer) []string {
	collector := map[string]struct{}{}
	set := []string{}
	lines := strings.Split(raw.String(), "\n")
	for _, line := range lines {
		name := strings.TrimSpace(line)
		if name != "" {
			collector[name] = struct{}{}
		}
	}
	for name := range collector {
		set = append(set, name)
	}
	return set
}
